#!/Users/christmas/opt/anaconda3/bin/python3
# -*- coding: utf-8 -*-
#  日期 : 2023/3/10 13:58
#  作者 : Christmas
#  邮箱 : 273519355@qq.com
#  项目 : Project
#  版本 : python 3
#  摘要 :
"""
processBar such as: ftp sftp
"""
import datetime
import time

import numpy as np

__all__ = ['translate_byte', 'LowSpeed_Error', 'SftpProcessbar', 'FtpProcessbar', 'FtpHtmlProcessbar']


def translate_byte(B):
    """
    :param B: bytes
    :return: 转换后的bytes
    """
    B = float(B)
    KB = float(1024)
    MB = float(KB * 1024)
    GB = float(MB * 1024)
    TB = float(GB * 1024)
    if B < KB:
        return f"{B} {'bytes' if B > 1 else 'byte'}"
    elif KB <= B < MB:
        return '{:.2f} KB'.format(B / KB)
    elif MB <= B < GB:
        return '{:.2f} MB'.format(B / MB)
    elif GB <= B < TB:
        return '{:.2f} GB'.format(B / GB)
    else:
        return '{:.2f} TB'.format(B / TB)


class LowSpeed_Error(Exception):
    """
    :param value: 低速异常
    """
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class SftpProcessbar(object):
    """
    sftp_obj =SftpProcessbar()
    Sprocess_bar = sftp_obj.process_bar

    # ssh = paramiko.SSHClient()
    # ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    # ssh.connect(hostname=host, port=22, username='wave', password='wave', timeout=100)
    # sftp = paramiko.SFTPClient.from_transport(ssh.get_transport())
    # sftp_obj =SftpProcessbar()
    # Sprocess_bar = sftp_obj.process_bar
    # sftp.put(local_path, a, callback=Sprocess_bar)
    """

    def __init__(self, bar_length=30, change_percent=1):
        self.bar_length = bar_length
        self.lastShownPercent = 0
        self.change_percent = change_percent

    def call_back(self, curr=100, total=100):
        bar_length = self.bar_length
        percents = '\033[32;1m%s\033[0m' % round(float(curr) * 100 / float(total), self.change_percent)
        filled = int(bar_length * curr / float(total))
        bar = '\033[32;1m%s\033[0m' % '=' * filled + '-' * (bar_length - filled)

        percentComplete = round((curr / total) * 100, self.change_percent)
        if self.lastShownPercent != percentComplete:
            self.lastShownPercent = percentComplete
            ddt = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
            print(f'{ddt} ---> [{bar}] {percents}% already complete: {translate_byte(curr)}, total: {translate_byte(total)}            \r', end='')

    def process_bar(self):
        """
        :return: call_back
        """
        return self.call_back


class FtpProcessbar(object):
    """
    ******已废弃******
    ******已废弃******
    ******已废弃******
    upload ------------------->:
    Ftp_obj = FtpProcessbar(os.path.getsize(local_path))
    Fprocess_bar = Ftp_obj.process_bar()

    # ftp = ftplib.FTP()
    # ftp.encoding = 'utf-8'
    # ftp.set_debuglevel(0)
    # ftp.connect(host=_host, port=_port)
    # ftp.login(_username, _password)
    # with open(local_path, 'rb') as fp:
    #   Ftp_obj = FtpProcessbar(os.path.getsize(local_path))
    #   Fprocess_bar = Ftp_obj.process_bar()
    #   ftp.storbinary(f'STOR {file}', fp, buf_size, Fprocess_bar)
    #   ==============================================
    download ------------------->:
    Ftp_obj = FtpProcessbar(ftp.size(local_path), fp=fp)
    Fprocess_bar = Ftp_obj.process_bar()

    # ftp = ftplib.FTP()
    # ftp.encoding = 'utf-8'
    # ftp.set_debuglevel(0)
    # ftp.connect(host=_host, port=_port)
    # ftp.login(_username, _password)
    # Ftp_obj = FtpProcessbar(ftp.size(local_path), fp=fp)
    # Fprocess_bar = Ftp_obj.process_bar()
    # with open(local_path, 'rb') as fp:
    #   Ftp_obj = FtpProcessbar(ftp.size(local_path), fp=fp)
    #   Fprocess_bar = Ftp_obj.process_bar()
    #   ftp.retrbinary(f'RETR {file}', Fprocess_bar,  buf_size)
    """

    def __init__(self, totalSize, fp=None, bar_length=30, change_percent=1):
        self.totalSize = totalSize  # 已上传或下载大小
        self.time_start = time.time()  # 开始上传或下载时间(注意是实例化时开始计时)
        self.sizeWritten = 0   # 文件总大小
        self.lastShownPercent = 0  # 上次显示的百分比
        self.bar_length = bar_length  # 进度条长度
        self.change_percent = change_percent  # 进度条变化百分比 0为1%输出 1为0.1%输出 2为0.01%输出
        if fp:
            self.fp = fp                   # 文件对象

    def call_back(self, block):
        """
        :param block: 上传或下载的数据块
        """
        if self.fp:
            self.fp.write(block)
        self.sizeWritten += len(block)
        percents = '\033[32;1m%s\033[0m' % round(float(self.sizeWritten) * 100 / float(self.totalSize), self.change_percent)
        filled = int(self.bar_length * self.sizeWritten / float(self.totalSize))
        bar = '\033[32;1m%s\033[0m' % '=' * filled + '-' * (self.bar_length - filled)
        speed = self.sizeWritten / (time.time() - self.time_start)

        percentComplete = round((self.sizeWritten / self.totalSize) * 100, self.change_percent)  # 粗分辨率,防止刷新过快
        if self.lastShownPercent != percentComplete:
            self.lastShownPercent = percentComplete
            ddt = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
            print(f'{ddt} ---> [{bar}] {percents}% speed:{translate_byte(speed)}/s, already complete: {translate_byte(self.sizeWritten)}, total: {translate_byte(self.totalSize)}            \r', end='')

    def process_bar(self):
        """
        :return: 返回一个函数对象
        """
        return self.call_back


class FtpHtmlProcessbar(object):
    """
    FTP_upload ------------------->:
    Ftp_obj = FtpHtmlProcessbar(ftp.size(local_path))
    Fprocess_bar = Ftp_obj.callback_ftp()

    ftp = ftplib.FTP()
    ftp.encoding = 'utf-8'
    ftp.set_debuglevel(0)
    ftp.connect(host=_host, port=_port)
    ftp.login(_username, _password)
    try:
        with open(local_path, 'rb') as fp:
            Ftp_obj = FtpHtmlProcessbar(os.path.getsize(local_path))
            Fprocess_bar = Ftp_obj.callback_ftp()
            ftp.storbinary(f'STOR {file}', fp, buf_size, Fprocess_bar)
    except LowSpeed_Error as e:
        print(e.value)
    #-----------------------------------------------#
    FTP_download ------------------->:
    Ftp_obj = FtpHtmlProcessbar(ftp.size(local_path), fp=fp)
    Fprocess_bar = Ftp_obj.callback_ftp()

    ftp = ftplib.FTP()
    ftp.encoding = 'utf-8'
    ftp.set_debuglevel(0)
    ftp.connect(host=_host, port=_port)
    ftp.login(_username, _password)
    try:
        with open(local_path, 'rb') as fp:
            Ftp_obj = FtpHtmlProcessbar(ftp.size(local_path), fp=fp)
            Fprocess_bar = Ftp_obj.callback_ftp()
            ftp.retrbinary(f'RETR {file}', Fprocess_bar,  buf_size)
    except LowSpeed_Error as e:
        print(e.value)

    #===============================================#
    HTML_download ------------------->:
    r = requests.get(url, stream=True)
    if r.status_code == 200:  # 请求成功
        chunksize = 102400    # 每次下载的数据大小
        try:
            with open(f'{self.grb_outpath}/{file}', "wb") as f:
                Html_obj = processBar.FtpHtmlProcessbar(int(r.headers['content-length']), fp=f, exit_num=20)  # 单位是byte
                for chunk in r.iter_content(chunksize):
                    if chunk:
                        Html_obj.callback_html(chunk)
        except LowSpeed_Error as e:
            print(e.value)
    """

    def __init__(self, totalSize, lowSpeed_floor=512*1024, fp=None, lowSpeed_exit_num=30, bar_length=30, change_percent=1, INFO=True):
        self.totalSize = totalSize                        # 文件总大小
        self.lowSpeed_floor = lowSpeed_floor              # 上传或下载退出速度(单位是byte),默认512KB/s
        self.time_start = np.float64(time.time())         # 开始上传或下载时间(注意是实例化时开始计时)
        self.sizeWritten = 0                              # 已上传或下载大小
        self.__lastShownPercent = 0                       # 上次显示的百分比
        self.lowSpeed_num = 0                             # 低速的次数
        self.lowSpeed_exit_num = lowSpeed_exit_num  # 退出次数
        self.__bar_length = bar_length                      # 进度条长度
        self.change_percent = change_percent              # 进度条变化百分比 0为1%输出 1为0.1%输出 2为0.01%输出
        self.INFO = INFO                                  # 是否显示进度条信息
        self.speed_list = []                              # 速度列表
        if fp:
            self.fp = fp                                  # 文件对象

    def callback_html(self, block):
        """
        :param block: 上传或下载的数据块
        """
        if self.fp:
            self.fp.write(block)
        self.sizeWritten += len(block)
        percents = '\033[32;1m%s\033[0m' % round(float(self.sizeWritten) * 100 / float(self.totalSize), self.change_percent)
        filled = int(self.__bar_length * self.sizeWritten / float(self.totalSize))
        bar = '\033[32;1m%s\033[0m' % '=' * filled + '-' * (self.__bar_length - filled)
        speed = self.sizeWritten / (np.float64(time.time()) - self.time_start)
        self.speed_list.append(speed)

        percentComplete = round((self.sizeWritten / self.totalSize) * 100, self.change_percent)  # 粗分辨率,防止刷新过快
        if self.__lastShownPercent != percentComplete:
            self.__lastShownPercent = percentComplete
            if self.INFO:
                ddt = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
                print('\r'+f'{ddt} ---> [{bar}] {percents}% speed:{translate_byte(speed)}/s, already complete: {translate_byte(self.sizeWritten)}, total: {translate_byte(self.totalSize)}', end='')
        if speed < self.lowSpeed_floor:
            self.lowSpeed_num += 1
            if self.lowSpeed_num >= self.lowSpeed_exit_num and percentComplete <= 85:
                raise LowSpeed_Error(speed)

    def callback_ftp(self):
        """
        :return: 返回一个函数对象
        """
        return self.callback_html



#!/Users/christmas/opt/anaconda3/bin/python3
# -*- coding: utf-8 -*-
#  日期 : 2025/10/16 14:13
#  作者 : Christmas
#  邮箱 : 273519355@qq.com
#  项目 : Project
#  版本 : python 3
#  摘要 :
"""

"""
from datetime import datetime
from typing import List, Union

import numpy as np
import pandas as pd
import xarray as xr

from .. import Ttime
from .xr_attrs import (add_lat_attrs, add_lon_attrs, add_time_attrs,
                       add_TIME_attrs, add_wave_attrs, fixed_encoding)


def xr_wave_dateset(_lon: np.ndarray,
                    _lat: np.ndarray,
                    _time: Union[List, str, np.ndarray, datetime, np.datetime64, pd.Timestamp],
                    **kwargs):
    """
    生成波浪数据的xarray.Dataset
    :param _lon: 经度
    :param _lat: 纬度
    :param _time: 时间
    :param kwargs: swh, mwd, mwp, mwl, pp1d, shww, shts, mdww, mdts, mpww, mpts, hmax,
                    engine, dtype, complevel, zlib, shuffle, fletcher32, contiguous,
                    product_name, source, start
        swh: significant height of combined wind waves and swell(m) 有效波高
        mwd: mean wave direction (deg) 平均波向
        mwp: mean wave period (s) 平均波周期
        mwl: mean wave length (m) 平均波长
        pp1d: peak period (s) 谱峰周期
        shww: significant height of wind waves (m) 风浪有效波高
        shts: significant height of total swell (m) 涌浪有效波高
        mdww: mean direction of wind waves (deg) 风浪波向
        mdts: mean direction of total swell (deg) 涌浪波向
        mpww: mean period of wind waves (s) 风浪平均周期
        mpts: mean period of total swell (s) 涌浪平均周期
        hmax: expected maximum wave height (linear, 1st order) (m) 最大波高
        engine: h5netcdf(default), netcdf4, scipy
        dtype: float32(default), float64
        complevel: 4(default) 0-9
        zlib: True(default)
        shuffle: True(default)
        fletcher32: True(default)
        contiguous: True(default)
        product_name: None(default) 产品名称
        source: Alaea(default) 数据来源
        start: None(default) 数据起始时间
    :return: xarray.Dataset
    """
    program_version = 'V1.0'
    history = 'Created by Python' + ' at ' + datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    if _time[0].__class__ == str:
        Ttimes = Ttime(TIME=_time)
    elif _time[0].__class__ == datetime or _time[0].__class__ == np.datetime64 or _time[0].__class__ == pd.Timestamp:
        Ttimes = Ttime(Times=_time)
    else:
        raise ValueError(_time[0].__class__)

    xr_wave = xr.Dataset(
        data_vars={
            "TIME": ("time", Ttimes.TIME)},
        coords={
            "longitude": _lon,
            "latitude": _lat,
            "time": Ttimes.time,
        },
        attrs={
            "product_name": kwargs.get('product_name', 'None'),
            "source": kwargs.get('source', 'Alaea'),
            "start": kwargs.get('start', 'None'),
            "history": history,
            "program_version": program_version,
        }
    )
    xr_wave = add_lon_attrs(xr_wave, lon_name='longitude')
    xr_wave = add_lat_attrs(xr_wave, lat_name='latitude')
    xr_wave = add_time_attrs(xr_wave, time_name='time')
    xr_wave = add_TIME_attrs(xr_wave, TIME_name='TIME', dim_name='DateStr')
    
    if 'swh' in kwargs:
        _swh = kwargs.get('swh')
        xr_wave['swh'] = (("time", "latitude", "longitude"), _swh)
    
    if 'mwd' in kwargs:
        _mwd = kwargs.get('mwd')
        xr_wave['mwd'] = (("time", "latitude", "longitude"), _mwd)
    
    if 'mwp' in kwargs:
        _mwp = kwargs.get('mwp')
        xr_wave['mwp'] = (("time", "latitude", "longitude"), _mwp)
    
    if 'mwl' in kwargs:
        _mwl = kwargs.get('mwl')
        xr_wave['mwl'] = (("time", "latitude", "longitude"), _mwl)
    
    if 'pp1d' in kwargs:
        _pp1d = kwargs.get('pp1d')
        xr_wave['pp1d'] = (("time", "latitude", "longitude"), _pp1d)
    
    if 'shww' in kwargs:
        _shww = kwargs.get('shww')
        xr_wave['shww'] = (("time", "latitude", "longitude"), _shww)
    
    if 'shts' in kwargs:
        _shts = kwargs.get('shts')
        xr_wave['shts'] = (("time", "latitude", "longitude"), _shts)
    
    if 'mdww' in kwargs:
        _mdww = kwargs.get('mdww')
        xr_wave['mdww'] = (("time", "latitude", "longitude"), _mdww)

    if 'mdts' in kwargs:
        _mdts = kwargs.get('mdts')
        xr_wave['mdts'] = (("time", "latitude", "longitude"), _mdts)
    
    if 'mpww' in kwargs:
        _mpww = kwargs.get('mpww')
        xr_wave['mpww'] = (("time", "latitude", "longitude"), _mpww)
    
    if 'mpts' in kwargs:
        _mpts = kwargs.get('mpts')
        xr_wave['mpts'] = (("time", "latitude", "longitude"), _mpts)
    
    if 'hmax' in kwargs:
        _hmax = kwargs.get('hmax')
        xr_wave['hmax'] = (("time", "latitude", "longitude"), _hmax)
    
    xr_wave = add_wave_attrs(xr_wave)
    
    # 去掉上面用过的key
    for key in ['swh', 'mwd', 'mwp', 'mwl', 'pp1d', 'shww', 'shts', 'mdww', 'mdts', 'mpww', 'mpts', 'hmax',
                'product_name', 'source', 'start']:
        if key in kwargs:
            kwargs.pop(key)
    xr_wave = fixed_encoding(xr_wave, **kwargs)
    
    # 排序
    ordered_vars = ['longitude', 'latitude', 'time', 'TIME',
                    'swh', 'mwd', 'mwp', 'mwl', 'pp1d', 'shww', 'shts', 'mdww', 'mdts', 'mpww', 'mpts', 'hmax']
    existing_vars = [var for var in ordered_vars if var in xr_wave.variables]
    xr_wave = xr_wave[existing_vars]

    return xr_wave



"""
https://pypi.org/project/Alaea/#history
"""
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="Alaea",
    version="2.3.4",
    author="Christmas",
    author_email="273519355@qq.com",
    description="Some simple tools by Christmas",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'numpy',
        'netCDF4',
        'xarray',
        'colorlog',
        'requests'
    ],
)




#!/Users/christmas/opt/anaconda3/bin/python3
# -*- coding: utf-8 -*-
#  日期 : 2025/1/10 15:10
#  作者 : Christmas
#  邮箱 : 273519355@qq.com
#  项目 : Project
#  版本 : python 3
#  摘要 :
"""
 
"""
from .fvcom_image import f_2d_image, f_2d_mesh
from .fvcom_load import f_load_grid, f_load_time

__all__ = ['FVCOM']


class FVCOM:
    def __init__(self,
                 fin,
                 Coordinate='geo',
                 MaxLon=180,
                 Global=False,
                 Nodisp=False):
        self.fin = fin
        self.grid = None
        self.Ttimes = None
        
        self.Coordinate = Coordinate
        self.MaxLon = MaxLon
        self.Global = Global
        self.Nodisp = Nodisp
        
        self.load()
    
    def load(self):
        self.grid = f_load_grid(self.fin, Coordinate=self.Coordinate, MaxLon=self.MaxLon, Global=self.Global, Nodisp=self.Nodisp)
        self.Ttimes = f_load_time(self.fin)

    def draw_2d_mesh(self, Color='k'):
        f_2d_mesh(self.grid, Global=self.Global, Color=Color)

    def draw_2d_image(self, var):
        f_2d_image(self.grid, var)


#!/Users/christmas/opt/anaconda3/bin/python3
# -*- coding: utf-8 -*-
#  日期 : 2025/1/9 13:36
#  作者 : Christmas
#  邮箱 : 273519355@qq.com
#  项目 : Project
#  版本 : python 3
#  摘要 :
"""
 
"""
import numpy as np

_all_ = ['calc_uv2sd', 'calc_uv2wind', 'calc_uv2current']


def calc_uv2sd(u, v, opt):
    """
    计算速度的速度方向和大小
    :param u: 东向速度
    :param v: 北向速度
    :param opt: 选项 wind current ww3
    :return: spd0, dir0
    """
    
    if opt == 'wind':
        spd0, dir0 = __uv2sd_to(u, v)
    elif opt == 'current':
        spd0, dir0 = __uv2sd_from(u, v)
    elif opt == 'ww3':
        spd0, dir0 = __uv2sd_from(u, v)
    else:
        raise ValueError(f"opt must be one of 'current','wind','wave','ww3', but you set '{opt}'")
    return spd0, dir0


def __uv2sd_from(u, v):
    """
    % uv to direction(from), such as wind, wave --> 来向 0 为正北，顺时针，90为正东
    :param u: 东向速度
    :param v: 北向速度
    :return: spd0, dir0
    """
    
    spd0, dir0 = calc_uv2wind(u, v)
    return spd0, dir0


def __uv2sd_to(u, v):
    """
    % uv to direction(to), such as current, --> 与矢量方向定义相同， 0为正东，逆时针，90为正北
    :param u: 东向速度
    :param v: 北向速度
    :return: spd0, dir0
    """
    
    spd0, dir0 = calc_uv2current(u, v)
    return spd0, dir0


def calc_uv2wind(u, v):
    """
    Calculate wind speed and direction from u, v
    :param u: 东向速度
    :param v: 北向速度
    :return: spd0, dir0
    """
    
    spd0 = np.sqrt(u ** 2 + v ** 2)
    dir0 = np.mod(270 - np.rad2deg(np.arctan2(v, u)), 360)
    return spd0, dir0


def calc_uv2current(u, v):
    """
    Calculate wind speed and direction from u, v
    :param u: 东向速度
    :param v: 北向速度
    :return: spd0, dir0
    """
    
    spd0 = np.sqrt(u ** 2 + v ** 2)
    dir0 = np.mod(np.rad2deg(np.arctan2(v, u)), 360)
    return spd0, dir0


#!/Users/christmas/opt/anaconda3/bin/python3
# -*- coding: utf-8 -*-
#  日期 : 2023/3/25 10:54
#  作者 : Christmas
#  邮箱 : 273519355@qq.com
#  项目 : Project
#  版本 : python 3
#  摘要 :
"""

"""
import datetime

import numpy as np
import pandas as pd

from . import convertToTime

__all__ = ['build_date', 'datetime_to_datenum', 'Times2Ttime', 'Ttime']


def build_date(_date):  # sourcery skip: inline-immediately-returned-variable
    """
    根据要跑模型的日期推出附近日期
    :param _date: 要跑模型的日期
    :return: 附近日期信息
    """
    today = str(_date)
    ymd_yes = (datetime.datetime.strptime(today, '%Y%m%d') - datetime.timedelta(days=1)).strftime('%Y%m%d')
    ymd_yes2 = (datetime.datetime.strptime(today, '%Y%m%d') - datetime.timedelta(days=2)).strftime('%Y%m%d')
    ymd_yes3 = (datetime.datetime.strptime(today, '%Y%m%d') - datetime.timedelta(days=3)).strftime('%Y%m%d')
    year_yes = ymd_yes[:4]
    month_yes = ymd_yes[4:6]
    year_yes2 = ymd_yes2[:4]
    month_yes2 = ymd_yes2[4:6]
    year_yes3 = ymd_yes3[:4]
    month_yes3 = ymd_yes3[4:6]

    ymd_today = datetime.datetime.strptime(today, '%Y%m%d').strftime('%Y%m%d')
    year_today = ymd_today[:4]
    month_today = ymd_today[4:6]

    ymd_tom = (datetime.datetime.strptime(today, '%Y%m%d') + datetime.timedelta(days=1)).strftime('%Y%m%d')
    ymd_tom2 = (datetime.datetime.strptime(today, '%Y%m%d') + datetime.timedelta(days=2)).strftime('%Y%m%d')
    ymd_tom3 = (datetime.datetime.strptime(today, '%Y%m%d') + datetime.timedelta(days=3)).strftime('%Y%m%d')
    year_tom = ymd_tom[:4]
    month_tom = ymd_tom[4:6]
    year_tom2 = ymd_tom2[:4]
    month_tom2 = ymd_tom2[4:6]
    year_tom3 = ymd_tom3[:4]
    month_tom3 = ymd_tom3[4:6]
    Nearby_date = {
        'ymd_yes': ymd_yes,
        'year_yes': year_yes,
        'month_yes': month_yes,
        'ymd_yes2': ymd_yes2,
        'year_yes2': year_yes2,
        'month_yes2': month_yes2,
        'ymd_yes3': ymd_yes3,
        'year_yes3': year_yes3,
        'month_yes3': month_yes3,

        'ymd_today': ymd_today,
        'year_today': year_today,
        'month_today': month_today,

        'ymd_tom': ymd_tom,
        'year_tom': year_tom,
        'month_tom': month_tom,
        'ymd_tom2': ymd_tom2,
        'year_tom2': year_tom2,
        'month_tom2': month_tom2,
        'ymd_tom3': ymd_tom3,
        'year_tom3': year_tom3,
        'month_tom3': month_tom3
    }
    return Nearby_date


def datetime_to_datenum(date):
    matlab_datenum_for_unix_epoch = 719529  # MATLAB 的 datenum 对于 1970-01-01 的值
    if not isinstance(date, list):
        days_from_epoch = (date - datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)).total_seconds() / (24 * 3600)  # Python 的 datetime 与 Unix 时间起点的差异（天）
        datenum = matlab_datenum_for_unix_epoch + days_from_epoch  # 计算 MATLAB datenum
    else:
        datenum = [datetime_to_datenum(date[i]) for i in range(len(date))]
    return datenum


def Times2Ttime(_Times):
    """
    将datetime格式的时间转换为Ttime格式
    :param _Times: datetime格式的时间
    :return: Ttime格式的时间
    """
    pass


class Ttime:
    def __init__(self, Times=None, TIME=None, fmt=None):
        """
        Times: datetime格式的时间
        TIME: 时间字符串
        fmt: 时间字符串的格式

        Updates:
        1. 2024-11-01, Fixed DeprecationWarning, by Christmas;
        2. 2025-10-16, Fixed tz_localize issue, by Christmas.
        """
        if TIME is not None and fmt is not None:
            if isinstance(TIME, list):
                self.Times = [datetime.datetime.strptime(TIME[i], fmt) for i in range(len(TIME))]
            else:
                self.Times = [datetime.datetime.strptime(TIME, fmt)]
        elif TIME is not None:
            if hasattr(TIME[0], 'dtype'):
                if isinstance(TIME, list):
                    TIME = ["".join(TIME[i].astype(str).tolist()) for i in range(len(TIME))]
                else:
                    TIME = ["".join(TIME.astype(str).tolist())]
            else:
                pass
            self.Times = convertToTime(listDate=TIME)
        else:
            if isinstance(Times[0], np.datetime64) or isinstance(Times[0], datetime.datetime):
                Times = [pd.to_datetime(Times[i]) for i in range(len(Times))]
            self.Times = Times

        # tz_localize UTC
        if isinstance(self.Times[0], datetime.datetime):
            self.Times = pd.to_datetime(self.Times)
        if self.Times[0].tzinfo is None:
            self.Times = [i.tz_localize('UTC') for i in self.Times]  # pd.datetime 才能用 tz_localize

        self.time = np.nan * np.zeros(len(self.Times))
        self.TIME = []
        self.TIME_str = [str() for _ in range(len(self.Times))]
        self.TIME_char = np.array([str() for _ in range(len(self.Times))])
        self.units = 'seconds since 1970-01-01 00:00:00'
        self.ymd = ''
        self.datenum = ''
        self.construct()

    def construct(self):
        # Times[i]     ---> datetime    || type: datetime.datetime
        # time[i]      ---> positive    || type: np.float64
        # TIME_str[i]  ---> str         || type: str
        # TIME_char[i] ---> char        || type: np.chararray
        # ymd[i]       ---> str         || type: str
        # datenum[i]   ---> float       || type: float
        self.time = [i.timestamp() for i in self.Times]
        self.TIME = [i.strftime('%Y-%m-%d %H:%M:%S') for i in self.Times]
        self.TIME_str = np.array(self.TIME)
        self.datenum = datetime_to_datenum(self.Times)

        # --> 2024-11-01, by Christmas
        # DeprecationWarning: `np.chararray` is deprecated and will be removed from the main namespace in the future. Use an array with a string or bytes dtype instead.
        max_length = max(len(t) for t in self.TIME)  # 找到最大字符串长度
        self.TIME_char = np.empty((len(self.TIME),), dtype='<U{}'.format(max_length))  # 创建字符串数组
        for i in range(len(self.TIME)):
            self.TIME_char[i] = self.TIME[i]
        self.ymd = [i.strftime('%Y%m%d') for i in self.Times]
        # -----------------------------------
        # char_array = np.chararray([len(self.TIME)], len(self.TIME[0]))
        # for i in range(len(self.TIME)):
        #     char_array[i] = self.TIME[i]
        # self.TIME_char = np.array(char_array)
        # self.ymd = [i.strftime('%Y%m%d') for i in self.Times]
        ## <--
